import pathlib
from typing import Union

try:
    import tomllib
except ModuleNotFoundError:
    # tomli backfills on Python < 3.11
    import tomli as tomllib

from schema import Schema, Optional, Or, Use, SchemaError, And

from tclint.violations import violation_types


class Config:
    def __init__(
        self,
        ignore=None,
        exclude=None,
        style_indent=None,
        style_line_length=None,
        style_aligned_set=None,
    ):
        self.ignore = []
        if ignore is not None:
            self.ignore = ignore

        self.exclude = []
        if exclude is not None:
            self.exclude = exclude

        self.style_indent = 4
        if style_indent is not None:
            self.style_indent = style_indent

        self.style_line_length = 80
        if style_line_length is not None:
            self.style_line_length = style_line_length

        self.style_aligned_set = False
        if style_aligned_set is not None:
            self.style_aligned_set = style_aligned_set

    @classmethod
    def from_dict(cls, config_dict: dict):
        schema = Schema({
            # note: it's ok if paths don't exist - allows for generic
            # configurations with directories like .git/ excluded
            Optional("exclude"): [Use(pathlib.Path)],
            Optional("ignore"): [
                Or(
                    And(
                        str,
                        lambda s: s in violation_types,
                        error="invalid rule ID provided for 'ignore'",
                    ),
                    {
                        "path": Use(pathlib.Path),
                        "rules": [
                            And(
                                str,
                                lambda s: s in violation_types,
                                error="invalid rule ID provided for 'ignore'",
                            )
                        ],
                    },
                )
            ],
            Optional("style"): {
                Optional("indent"): Or(
                    lambda v: v == "tab", int, error="indent must be integer or 'tab'"
                ),
                Optional("line-length"): Use(int, error="line-length must be integer"),
                Optional("allow-aligned-sets"): Use(
                    bool, error="allow-aligned-sets must be a bool"
                ),
            },
        })

        try:
            config = schema.validate(config_dict)
        except SchemaError as e:
            error_s = str(e).replace("\n", " ")
            raise ConfigError(error_s)

        style_config = config.get("style", {})

        return cls(
            exclude=config.get("exclude", None),
            ignore=config.get("ignore", None),
            style_indent=style_config.get("indent", None),
            style_line_length=style_config.get("line-length", None),
            style_aligned_set=style_config.get("allow-aligned-sets", None),
        )

    @classmethod
    def from_path(cls, path: Union[str, pathlib.Path]):
        path = pathlib.Path(path)

        if not path.exists():
            raise FileNotFoundError

        with open(path, "rb") as f:
            try:
                data = tomllib.load(f)
            except tomllib.TOMLDecodeError as e:
                raise ConfigError(f"{path}: {e}")

        try:
            return cls.from_dict(data)
        except ConfigError as e:
            raise ConfigError(f"{path}: {e}")

    @classmethod
    def from_pyproject(cls):
        path = pathlib.Path("pyproject.toml")

        if not path.exists():
            raise FileNotFoundError

        with open("pyproject.toml", "r") as f:
            data = tomllib.load(f)

        tclint_config = data.get("tool", {})["tclint"]

        try:
            return cls.from_dict(tclint_config)
        except ConfigError as e:
            raise ConfigError(f"pyproject.toml: {e}")


class ConfigError(Exception):
    pass


def get_config(config_path=None) -> Config:
    DEFAULT_CONFIGS = ("tclint.toml", ".tclint")

    # user-supplied
    if config_path is not None:
        try:
            return Config.from_path(config_path)
        except FileNotFoundError:
            raise ConfigError(f"path {config_path} doesn't exist")

    for path in DEFAULT_CONFIGS:
        try:
            return Config.from_path(path)
        except FileNotFoundError:
            pass

    try:
        return Config.from_pyproject()
    except ConfigError as e:
        raise e
    except Exception:
        pass

    return Config()



from tclint.lexer import (
    Lexer,
    TclSyntaxError,
    TOK_BACKSLASH_NEWLINE,
    TOK_NEWLINE,
    TOK_SEMI,
    TOK_WS,
    TOK_QUOTE,
    TOK_ARG_EXPANSION,
    TOK_LBRACE,
    TOK_RBRACE,
    TOK_LBRACKET,
    TOK_RBRACKET,
    TOK_DOLLAR,
    TOK_LPAREN,
    TOK_RPAREN,
    TOK_HASH,
    TOK_VAR_CHARS,
    TOK_NAMESPACE_SEP,
    TOK_EOF,
)
from tclint.syntax_tree import (
    Script,
    Comment,
    Command,
    CommandSub,
    ArgExpansion,
    VarSub,
    BareWord,
    BracedWord,
    QuotedWord,
    CompoundBareWord,
    List,
    Expression,
)
from tclint.commands.builtin import CommandArgError, commands
from tclint.violations import CommandArgViolation


class _Word:
    """Helper class for constructing Word nodes out of multiple segments."""

    def __init__(self):
        self.segments = []
        self.current_segment = ""
        self.current_start = None

    def add_tok(self, tok):
        if self.current_start is None:
            self.current_start = tok.value[1]
        self.current_segment += tok.value[0]

    def add_node(self, node):
        if self.current_segment != "":
            self.segments.append(
                BareWord(self.current_segment, pos=self.current_start, end_pos=node.pos)
            )
            self.current_segment = ""
            self.current_start = None
        self.segments.append(node)

    def resolve(self, end_pos):
        if self.current_segment:
            self.segments.append(
                BareWord(self.current_segment, pos=self.current_start, end_pos=end_pos)
            )

        return self.segments


class Parser:
    def __init__(self, debug=False, cmd_sub=False, debug_indent=0):
        """cmd_sub = True indicates that the script will be terminated by ]"""
        self._debug = debug
        self._debug_indent = debug_indent
        self._cmd_sub = cmd_sub
        # TODO: better way to handle this?
        self.violations = []

    def debug(self, *msg):
        if self._debug:
            print("  " * self._debug_indent, end="")
            print(*msg)

    def parse(self, script, pos=None):
        lexer = Lexer(pos=pos)
        lexer.input(script)
        tree = self.parse_script(lexer)

        return tree

    def _parse_command_args(self, routine, args):
        """Since many built-in Tcl commands take in Tcl scripts or expressions
        as arguments, building a complete parse tree requires checking command
        names and possibly parsing their arguments.

        The node of any argument that gets parsed by this method is replaced with
        the parse tree of that argument. Since this process requires checking
        the arguments provided to these commands, this method may report lint
        violations.

        This parsing process is analogous to how the Tcl interpreter interprets
        scripts, and better handles weird edge cases compared to a traditional
        parsing technique. For example, this may look like valid Tcl:

        proc foo {a} {
            # output }
            puts "}"
        }

        But really, it is invalid since the } in the comment terminates the body
        of the proc - Tcl blindly constructs the body of the proc until it
        reaches the first }. tclint handles this correctly.
        """
        if routine not in commands:
            return args

        func = commands[routine]
        if func is None:
            return args

        try:
            new_args = func(args, self)
        except CommandArgError as e:
            raise e
        except Exception:
            if self.debug:
                raise
            raise CommandArgError(
                f"error parsing command arguments, possibly malformed {routine} command"
            )

        if new_args is None:
            return args

        # This constraint ensures the parse tree represents the shape of the
        # code for style checking purposes (e.g. arg spacing)
        assert len(new_args) == len(args)

        return new_args

    def parse_script(self, ts):
        self.debug(f"parse_script({ts.current})")
        self._debug_indent += 1
        pos = ts.pos()

        if self._cmd_sub:
            script = CommandSub(pos=pos)
        else:
            script = Script(pos=pos)

        while ts.type() is not TOK_EOF:
            if ts.type() in {TOK_WS, TOK_BACKSLASH_NEWLINE}:
                # strip whitespace at start of command
                ts.next()
                continue

            if ts.type() == TOK_HASH:
                script.add(self.parse_comment(ts))
            else:
                cmd = self.parse_command(ts)
                if cmd is not None:
                    script.add(cmd)

            # when in command sub mode, a script is terminated by ]
            if self._cmd_sub and ts.type() == TOK_RBRACKET:
                return script

            ts.expect(
                TOK_EOF,
                TOK_NEWLINE,
                TOK_SEMI,
                message=(
                    f"Expected newline or semicolon at {ts.pos()}, got {ts.value()}"
                ),
            )

        if self._cmd_sub and ts.type() is TOK_EOF:
            raise TclSyntaxError(
                "reached EOF without finding end of command substitution starting at"
                f" {pos}"
            )

        self._debug_indent -= 1

        script.end_pos = ts.pos()

        return script

    def parse_comment(self, ts):
        self.debug(f"parse_comment({ts.current})")
        pos = ts.pos()

        ts.assert_(TOK_HASH)

        value = ""
        while ts.type() not in {TOK_NEWLINE, TOK_EOF}:
            value += ts.value()
            ts.next()

        return Comment(value, pos=pos, end_pos=ts.pos())

    def parse_command(self, ts):
        self.debug(f"parse_command({ts.current})")
        self._debug_indent += 1
        pos = ts.pos()

        routine = self.parse_word(ts)
        if routine is None:
            self._debug_indent -= 1
            return None

        args = []
        while True:
            if ts.type() not in {TOK_WS, TOK_BACKSLASH_NEWLINE}:
                break

            while ts.type() in {TOK_WS, TOK_BACKSLASH_NEWLINE}:
                ts.next()

            word = self.parse_word(ts)
            if word is None:
                break

            args.append(word)

        self._debug_indent -= 1

        try:
            parsed_args = self._parse_command_args(routine.contents, args)
        except CommandArgError as e:
            self.violations.append(CommandArgViolation(str(e), pos))
            parsed_args = args

        return Command(routine, *parsed_args, pos=pos, end_pos=ts.pos())

    def parse_word(self, ts):
        self.debug(f"parse_word({ts.current})")
        if ts.type() == TOK_ARG_EXPANSION:
            return self.parse_arg_expansion(ts)
        elif ts.type() == TOK_LBRACE:
            return self.parse_braced_word(ts)
        elif ts.type() == TOK_QUOTE:
            return self.parse_quoted_word(ts)
        else:
            return self.parse_bare_word(ts)

    def parse_arg_expansion(self, ts):
        self.debug(f"parse_arg_expansion({ts.current})")
        pos = ts.pos()

        ts.assert_(TOK_ARG_EXPANSION)

        # Arg expansion is just a regular braced word if followed by whitespace
        if ts.type() in {TOK_WS, TOK_BACKSLASH_NEWLINE, TOK_NEWLINE, TOK_EOF}:
            return BracedWord("*", pos=pos, end_pos=ts.pos())

        return ArgExpansion(self.parse_word(ts), pos=pos, end_pos=ts.pos())

    def parse_quoted_word(self, ts):
        self.debug(f"parse_quoted_word({ts.current})")
        self._debug_indent += 1
        pos = ts.pos()

        ts.assert_(TOK_QUOTE)

        word = _Word()
        while ts.type() not in {TOK_QUOTE, TOK_EOF}:
            if ts.type() == TOK_DOLLAR:
                dollar_tok = ts.current
                var_sub = self.parse_var_sub(ts)
                if var_sub:
                    word.add_node(var_sub)
                else:
                    word.add_tok(dollar_tok)
            elif ts.type() == TOK_LBRACKET:
                command_sub = self.parse_command_sub(ts)
                word.add_node(command_sub)
            else:
                word.add_tok(ts.current)
                ts.next()

        if ts.type() == TOK_EOF:
            raise TclSyntaxError(
                f"reached EOF without finding match for quote at {pos}"
            )

        res = word.resolve(ts.pos())

        ts.assert_(TOK_QUOTE)

        self._debug_indent -= 1

        if not res:
            res = []

        return QuotedWord(*res, pos=pos, end_pos=ts.pos())

    def parse_braced_word(self, ts):
        self.debug(f"parse_braced_word({ts.current})")
        pos = ts.pos()

        ts.assert_(TOK_LBRACE)

        word = ""
        # store position for each brace we want to match, facilitating good
        # error messages
        expected_braces = [pos]
        while True:
            if ts.type() == TOK_EOF:
                raise TclSyntaxError(
                    "Reached EOF without finding match for brace at"
                    f" {expected_braces[-1]}"
                )

            if ts.type() == TOK_LBRACE:
                expected_braces.append(ts.pos())
            elif ts.type() == TOK_RBRACE:
                try:
                    expected_braces.pop()
                except IndexError:
                    raise TclSyntaxError(
                        f"Found closing brace at {ts.pos()} without matching open brace"
                    )

                if len(expected_braces) == 0:
                    ts.next()
                    break
            word += ts.value()
            ts.next()

        end_pos = ts.pos()
        return BracedWord(word, pos=pos, end_pos=end_pos)

    def parse_bare_word(self, ts):
        self.debug(f"parse_bare_word({ts.current})")
        self._debug_indent += 1
        pos = ts.pos()

        word = _Word()
        delimiters = [TOK_WS, TOK_BACKSLASH_NEWLINE, TOK_NEWLINE, TOK_SEMI, TOK_EOF]

        # In command sub mode, words are ended by ]
        if self._cmd_sub:
            delimiters.append(TOK_RBRACKET)

        while ts.type() not in delimiters:
            if ts.type() == TOK_DOLLAR:
                dollar_tok = ts.current
                var_sub = self.parse_var_sub(ts)
                if var_sub:
                    word.add_node(var_sub)
                else:
                    word.add_tok(dollar_tok)
            elif ts.type() == TOK_LBRACKET:
                command_sub = self.parse_command_sub(ts)
                word.add_node(command_sub)
            else:
                word.add_tok(ts.current)
                ts.next()

        res = word.resolve(ts.pos())

        self._debug_indent -= 1

        if not res:
            return None
        if len(res) == 1:
            return res[0]
        return CompoundBareWord(*res, pos=pos, end_pos=ts.pos())

    def parse_var_sub(self, ts):
        self.debug(f"parse_var_sub({ts.current})")
        pos = ts.pos()

        ts.assert_(TOK_DOLLAR)

        var = ""
        if ts.type() == TOK_LBRACE:
            brace_pos = ts.pos()
            ts.next()
            while ts.type() != TOK_RBRACE:
                if ts.type() is TOK_EOF:
                    raise TclSyntaxError(
                        f"Reached EOF without finding match for brace at {brace_pos}"
                    )
                var += ts.value()
                ts.next()
            ts.next()

            return VarSub(var, pos=pos, end_pos=ts.pos())

        while ts.type() in {TOK_VAR_CHARS, TOK_NAMESPACE_SEP}:
            var += ts.value()
            ts.next()

        if not var:
            return None

        index_nodes = []
        if ts.type() == TOK_LPAREN:
            paren_pos = ts.pos()
            index = _Word()
            ts.next()
            while ts.type() != TOK_RPAREN:
                if ts.type() == TOK_EOF:
                    raise TclSyntaxError(
                        f"Reached EOF without finding match for paren at {paren_pos}"
                    )
                if ts.type() == TOK_DOLLAR:
                    dollar_tok = ts.current
                    var_sub = self.parse_var_sub(ts)
                    if var_sub:
                        index.add_node(var_sub)
                    else:
                        index.add_tok(dollar_tok)
                elif ts.type() == TOK_LBRACKET:
                    command_sub = self.parse_command_sub(ts)
                    index.add_node(command_sub)
                else:
                    index.add_tok(ts.current)
                    ts.next()

            index_nodes = index.resolve(ts.pos())
            ts.next()

        var_sub = VarSub(var, pos=pos, end_pos=ts.pos())

        for index_segment in index_nodes:
            var_sub.add(index_segment)

        return var_sub

    def parse_command_sub(self, ts):
        self.debug(f"parse_command_sub({ts.current})")
        self._debug_indent += 1

        pos = ts.pos()
        ts.assert_(TOK_LBRACKET)

        saved_cmd_sub = self._cmd_sub
        self._cmd_sub = True
        script = self.parse_script(ts)
        self._cmd_sub = saved_cmd_sub

        ts.assert_(TOK_RBRACKET)
        end_pos = ts.pos()

        script.line = pos[0]
        script.col = pos[1]
        script.end_pos = end_pos

        self._debug_indent -= 1
        return script

    def parse_list(self, node):
        """Parse contents of node as Tcl list. This is a distinct entry point
        that doesn't get used when generating the main syntax tree, but is used
        in command-specific argument parsing.
        """
        if node.contents is None:
            return None

        # Hack: we know contents start 1 col over, since first char denotes the list
        ts = Lexer(pos=(node.line, node.col + 1))
        ts.input(node.contents)

        DELIMITERS = {TOK_WS, TOK_BACKSLASH_NEWLINE, TOK_NEWLINE}

        list_node = List(pos=node.pos, end_pos=node.end_pos)
        while ts.type() is not TOK_EOF:
            while ts.type() in DELIMITERS:
                ts.next()

            if ts.type() is TOK_EOF:
                break

            if ts.type() == TOK_LBRACE:
                # we can reuse parse_braced_word, since it doesn't use
                # substitutions in any case
                list_node.add(self.parse_braced_word(ts))
            elif ts.type() == TOK_QUOTE:
                quote_word_pos = ts.pos()

                ts.assert_(TOK_QUOTE)

                bare_word_pos = ts.pos()
                contents = ""
                while ts.type() not in {TOK_QUOTE, TOK_EOF}:
                    contents += ts.value()
                    ts.next()
                word = BareWord(contents, pos=bare_word_pos, end_pos=ts.pos())

                ts.assert_(TOK_QUOTE)

                list_node.add(QuotedWord(word, pos=quote_word_pos, end_pos=ts.pos()))
            else:
                pos = ts.pos()
                contents = ""
                while ts.type() not in DELIMITERS:
                    contents += ts.value()
                    ts.next()
                list_node.add(BareWord(contents, pos=pos, end_pos=ts.pos()))

        return list_node

    def parse_expression(self, node):
        return Expression(node.contents, pos=node.pos, end_pos=node.end_pos)



from typing import Tuple


class Violation:
    def __init__(self, id: str, message: str, pos: Tuple[int, int]):
        self.id = id
        self.message = message
        self.pos = pos

    def __lt__(self, other):
        return self.pos < other.pos

    def __str__(self):
        line, col = self.pos
        return f"{line}:{col}: {self.message} [{self.id}]"

    @classmethod
    def create(cls, id):
        def func(message: str, pos: Tuple[int, int]):
            return cls(id, message, pos)

        return func


violation_types = [
    "indent",
    "spacing",
    "line-length",
    "trailing-whitespace",
    "command-args",
]

IndentViolation = Violation.create("indent")
SpacingViolation = Violation.create("spacing")
LineLengthViolation = Violation.create("line-length")
TrailingWhiteSpaceViolation = Violation.create("trailing-whitespace")
# used by parser. TODO: should this be separated per command?
CommandArgViolation = Violation.create("command-args")



import pathlib

import pytest

from tclint.config import get_config, Config, ConfigError

MY_DIR = pathlib.Path(__file__).parent.resolve()


def test_example_config():
    config_path = MY_DIR / "data" / "tclint.toml"
    config = get_config(config_path)

    assert config.exclude == list(map(pathlib.Path, ["ignore_me/", "ignore.tcl"]))
    assert config.ignore == [
        "spacing",
        {"path": pathlib.Path("files_with_bad_indent/"), "rules": ["indent"]},
    ]

    assert config.style_indent == 2
    assert config.style_line_length == 100
    assert config.style_aligned_set is True


def test_invalid_rule():
    with pytest.raises(ConfigError):
        Config.from_dict({"ignore": ["asdf"]})



def is_relative_to(this, other):
    """Can be replaced by pathlib.Path.is_relative_to() once we drop Python 3.8."""
    try:
        this.relative_to(other)
        return True
    except ValueError:
        return False



import pathlib
import subprocess

import pytest

MY_DIR = pathlib.Path(__file__).parent.resolve()

tests = []
test_case_dir = MY_DIR / "data"
for path in test_case_dir.iterdir():
    if path.suffix == ".tcl":
        tests.append(str(path.relative_to(MY_DIR)))


@pytest.mark.parametrize("test", tests)
def test_tclint(test):
    """End-to-end tests."""
    output_path = (MY_DIR / test).with_suffix(".txt")
    with open(output_path, "r") as f:
        expected = f.read()

    p = subprocess.run(["tclint", test], capture_output=True, cwd=MY_DIR)

    assert p.stdout.decode("utf-8") == expected
    assert p.returncode == 0 if not expected else 1



import re

from tclint.violations import (
    IndentViolation,
    SpacingViolation,
    TrailingWhiteSpaceViolation,
    LineLengthViolation,
)
from tclint.syntax_tree import Visitor, Script, List


class IndentLevelChecker(Visitor):
    """Checks that each line is indented at the correct level.

    Reports 'indent' violations.
    """

    def __init__(self):
        # -1 lets us add 1 in all script visitors and get the correct level in
        # the top-level script
        self.level = -1

        # maps lines to expected indent levels
        self.expected_levels = {}

    def check(self, input, tree, config):
        # first have visitor calculate expected indent on each line
        tree.accept(self, recurse=False)

        # then iterate line-by-line to verify that actual indentation matches
        violations = []
        for lineno, line in enumerate(input.split("\n")):
            # 1-indexed
            lineno += 1

            if lineno not in self.expected_levels:
                continue

            expected_level = self.expected_levels[lineno]
            indent = "\t" if config.style_indent == "tab" else " " * config.style_indent
            expected_indent = indent * expected_level

            actual_indent = ""
            for c in line:
                if c in {" ", "\t"}:
                    actual_indent += c
                else:
                    break

            if expected_indent != actual_indent:
                if config.style_indent == "tab":
                    expected_str = (
                        f"{expected_level} tab{'s' if expected_level != 1 else ''}"
                    )
                else:
                    count = expected_level * config.style_indent
                    expected_str = f"{count} space{'s' if count != 1 else ''}"

                if " " in actual_indent and "\t" in actual_indent:
                    actual_str = "mix of tabs and spaces"
                elif "\t" in actual_indent:
                    actual_str = f"{len(actual_indent)}"
                    if config.style_indent != "tab":
                        actual_str += f" tab{'s' if len(actual_indent) != 1 else ''}"
                else:
                    actual_str = f"{len(actual_indent)}"
                    if config.style_indent == "tab":
                        actual_str += f" space{'s' if len(actual_indent) != 1 else ''}"

                pos = (lineno, 1)
                violations.append(
                    IndentViolation(
                        f"expected indent of {expected_str}, got {actual_str}",
                        pos,
                    )
                )

        return violations

    def visit_script(self, script):
        self._visit_script(script)

    def visit_list(self, list):
        # List works like Script for indent checking purposes
        self._visit_script(list)

    def _visit_script(self, script):
        self.level += 1
        for child in script.children:
            child.accept(self)
        self.level -= 1

        # check indentation of closing token of script
        if script.end_pos:
            end_line, _ = script.end_pos
            if end_line not in self.expected_levels:
                self.expected_levels[end_line] = self.level

    def visit_command(self, command):
        assert len(command.children) > 0

        if command.line in self.expected_levels:
            # already visited something on this line
            # TODO: do we want to return here? do we still wanna check continuation?
            return

        self.expected_levels[command.line] = self.level

        for child in command.children:
            # continuations of command must be indented. If the node is a list
            # or script however, the indentation only applies to contents of that node
            # (handled by the relevant visitors)
            if not isinstance(child, (Script, List)):
                self.level += 1

            child.accept(self)

            if not isinstance(child, (Script, List)):
                self.level -= 1

    def visit_comment(self, comment):
        self._set_level(comment)

    def visit_command_sub(self, command_sub):
        self._set_level(command_sub)

    def visit_bare_word(self, word):
        self._set_level(word)

    def visit_braced_word(self, word):
        self._set_level(word)

    def visit_quoted_word(self, word):
        self._set_level(word)

    def visit_compound_bare_word(self, word):
        self._set_level(word)

    def visit_var_sub(self, var_sub):
        self._set_level(var_sub)

    def visit_arg_expansion(self, arg_expansion):
        self._set_level(arg_expansion)

    def _set_level(self, node):
        if node.line in self.expected_levels:
            return

        self.expected_levels[node.line] = self.level


class SpacingChecker(Visitor):
    """Ensures each word of a command is separated by a single space (if on the
    same line).

    Reports 'spacing' violations.
    """

    def __init__(self):
        self.violations = []
        self.set_groups = [
            []
            # [set cmd...]
        ]

    def check(self, _, tree, config):
        self.aligned_set = config.style_aligned_set

        tree.accept(self, recurse=True)

        if self.aligned_set:
            self._check_set_spacing()

        return self.violations

    def visit_command(self, command):
        if len(command.args) == 0:
            return

        handle_as_aligned_set = command.routine == "set" and self.aligned_set

        last_word = command.children[0]
        for i, word in enumerate(command.children[1:]):
            if last_word.end_pos[0] != word.pos[0]:
                # ignore if on diff lines
                continue

            if handle_as_aligned_set and i == 1:
                # relies on visit_command being called in order
                if (
                    len(self.set_groups[-1]) > 0
                    and command.line == self.set_groups[-1][-1].line + 1
                ):
                    self.set_groups[-1].append(command)
                else:
                    self.set_groups.append([command])

                continue

            spacing = word.pos[1] - last_word.end_pos[1]
            if spacing != 1:
                self.violations.append(
                    SpacingViolation(
                        f"expected 1 space between words, got {spacing}",
                        last_word.end_pos,
                    )
                )
            last_word = word

    def _check_set_spacing(self):
        for group in self.set_groups:
            furthest_value_start = 0
            for i, cmd in enumerate(group):
                value_start = cmd.args[1].pos[1]
                if value_start > furthest_value_start:
                    furthest_value_start = value_start

            violations = []
            all_aligned = True
            for i, cmd in enumerate(group):
                if cmd.args[1].pos[1] != furthest_value_start:
                    all_aligned = False

                spacing = cmd.args[1].pos[1] - cmd.args[0].end_pos[1]
                if spacing != 1:
                    violations.append(
                        SpacingViolation(
                            "expected 1 space between value and name or fully aligned"
                            " block",
                            cmd.args[1].pos,
                        )
                    )

            # If all set values are aligned, waive violations. Alignment is
            # meaningless if there's only one command in the group, always
            # report in that case.
            if len(group) == 1 or not all_aligned:
                self.violations.extend(violations)


class LineChecker:
    """Ensures lines aren't too long and do not include trailing whitespace.

    Reports 'line-length' and 'trailing-whitespace' violations.
    """

    # ref: https://github.com/eslint/eslint/blob/b29a16b22f234f6134475efb6c7be5ac946556ee/lib/rules/max-len.js#L101 # noqa: E501
    # ^ ironic lint waiver...
    URL_RE = re.compile(r"[^:/?#]:\/\/[^?#]")

    def check(self, input, _, config):
        violations = []
        for i, line in enumerate(input.split("\n")):
            if self.URL_RE.search(line) is not None:
                # ignore URLs
                continue

            lineno = i + 1
            if len(line) > config.style_line_length:
                pos = (lineno, config.style_line_length + 1)
                violations.append(
                    LineLengthViolation(
                        f"line length is {len(line)}, maximum allowed is"
                        f" {config.style_line_length}",
                        pos,
                    )
                )

            if line.endswith((" ", "\t")):
                pos = (lineno, len(line))
                violations.append(
                    TrailingWhiteSpaceViolation("line has trailing whitespace", pos)
                )

        return violations



import ply.lex as lex

TOK_BACKSLASH_NEWLINE = "BACKSLASH_NEWLINE"
TOK_BACKSLASH_SUB = "BACKSLASH_SUB"
TOK_NEWLINE = "NEWLINE"
TOK_SEMI = "SEMI"
TOK_WS = "WS"
TOK_QUOTE = "QUOTE"
TOK_ARG_EXPANSION = "ARG_EXPANSION"
TOK_LBRACE = "LBRACE"
TOK_RBRACE = "RBRACE"
TOK_STAR = "STAR"
TOK_LBRACKET = "LBRACKET"
TOK_RBRACKET = "RBRACKET"
TOK_DOLLAR = "DOLLAR"
TOK_LPAREN = "LPAREN"
TOK_RPAREN = "RPAREN"
TOK_HASH = "HASH"
TOK_VAR_CHARS = "VAR_CHARS"
TOK_NAMESPACE_SEP = "NAMESPACE_SEP"
TOK_CHAR = "CHAR"
TOK_EOF = None


class TclSyntaxError(Exception):
    pass


class Lexer:
    tokens = (
        TOK_BACKSLASH_NEWLINE,
        TOK_BACKSLASH_SUB,
        TOK_NEWLINE,
        TOK_SEMI,
        TOK_WS,
        TOK_QUOTE,
        TOK_ARG_EXPANSION,
        TOK_LBRACE,
        TOK_RBRACE,
        TOK_STAR,
        TOK_LBRACKET,
        TOK_RBRACKET,
        TOK_DOLLAR,
        TOK_LPAREN,
        TOK_RPAREN,
        TOK_HASH,
        TOK_VAR_CHARS,
        TOK_NAMESPACE_SEP,
        TOK_CHAR,
    )

    def _tok(self, t):
        pos = (t.lexer.lineno, self.colno)
        if "\n" in t.value:
            t.lexer.lineno += t.value.count("\n")
            assert t.value[-1] == "\n"
            self.colno = 1
        else:
            self.colno += len(t.value)
        t.value = (t.value, pos)
        return t

    # Priority important
    def t_BACKSLASH_NEWLINE(self, t):
        r"\\\n"
        return self._tok(t)

    # Priority important
    def t_BACKSLASH_SUB(self, t):
        r"\\."
        return self._tok(t)

    def t_NEWLINE(self, t):
        r"\n"
        return self._tok(t)

    def t_SEMI(self, t):
        r";"
        return self._tok(t)

    # TODO: should use \s?
    def t_WS(self, t):
        r"[\t\v\f\r ]+"
        return self._tok(t)

    def t_QUOTE(self, t):
        r'"'
        return self._tok(t)

    # Must be higher priority than LBRACE
    def t_ARG_EXPANSION(self, t):
        r"\{\*\}"
        return self._tok(t)

    def t_LBRACE(self, t):
        r"\{"
        return self._tok(t)

    def t_RBRACE(self, t):
        r"\}"
        return self._tok(t)

    def t_STAR(self, t):
        r"\*"
        return self._tok(t)

    def t_LBRACKET(self, t):
        r"\["
        return self._tok(t)

    def t_RBRACKET(self, t):
        r"\]"
        return self._tok(t)

    def t_DOLLAR(self, t):
        r"\$"
        return self._tok(t)

    def t_LPAREN(self, t):
        r"\("
        return self._tok(t)

    def t_RPAREN(self, t):
        r"\)"
        return self._tok(t)

    def t_HASH(self, t):
        r"\#"
        return self._tok(t)

    # Valid chars in variable names
    def t_VAR_CHARS(self, t):
        r"[A-Za-z0-9_]+"
        return self._tok(t)

    def t_NAMESPACE_SEP(self, t):
        r"::+"
        return self._tok(t)

    # Catch-all. TODO: inefficient, should probably munch multiple chars
    def t_CHAR(self, t):
        r"."
        return self._tok(t)

    # Error handling rule
    # TODO: do we need this? since we have a catch-all...
    # there is a warning
    def t_error(self, t):
        print("Illegal character '%s'" % t.value[0])
        t.lexer.skip(1)

    def __init__(self, pos=None):
        self.lexer = lex.lex(object=self)
        self.current = None
        self.colno = 1

        if pos is not None:
            line, col = pos
            # use built-in lineno since it aids debugging, must store colno
            # separately since lex doesn't track it
            self.lexer.lineno = line
            self.colno = col

    def input(self, text):
        self.lexer.input(text)
        self.current = self.lexer.token()

    def type(self):
        if self.current is None:
            return TOK_EOF
        return self.current.type

    def value(self):
        if self.current is None:
            return None
        return self.current.value[0]

    def pos(self):
        if self.current is None:
            return (self.lexer.lineno, self.colno)
        return self.current.value[1]

    def next(self):
        self.current = self.lexer.token()

    def expect(self, *tokens, message=None):
        if self.type() not in tokens:
            raise TclSyntaxError(message)

        self.next()

    def assert_(self, *tokens):
        assert self.current.type in tokens
        self.next()



from pathlib import Path

from tclint.tclint import lint
from tclint.config import Config


def test_cmd_args_in_sub():
    """Ensure command args get checked in cmd sub."""
    script = "[puts]"
    violations = lint(script, Config(), Path())

    assert len(violations) == 1
    assert violations[0].id == "command-args"


def test_no_false_positive_arg_expansion():
    """Ensure that argument count checks don't flag when the number of arguments
    is ambiguous."""
    # $foo may be a list with two items, which is legal
    script = "rename {*}$foo"
    violations = lint(script, Config(), Path())
    assert len(violations) == 0


def test_aligned_sets():
    """Test style.allow-aligned-sets"""
    script = """
set foo  0
set barx 1"""

    violations = lint(script, Config(style_aligned_set=False), Path())
    assert len(violations) == 1
    assert violations[0].id == "spacing"

    violations = lint(script, Config(style_aligned_set=True), Path())
    assert len(violations) == 0


def test_accidental_return_expr():
    script = "return 5 + 2"
    violations = lint(script, Config(), Path())
    assert len(violations) == 1
    assert violations[0].id == "command-args"


def test_ignore_path():
    script = "puts  bad_spacing"
    fake_path = Path("bad.tcl")
    violations = lint(script, Config(), fake_path)
    assert len(violations) == 1

    config = Config(ignore=[{"path": fake_path, "rules": ["spacing"]}])
    violations = lint(script, config, fake_path)
    assert len(violations) == 0



"""Main CLI entry point."""

import argparse
import codecs
import os
import pathlib
import sys
from typing import List

from tclint.config import get_config, ConfigError, Config
from tclint.parser import Parser, TclSyntaxError
from tclint.checks import IndentLevelChecker, SpacingChecker, LineChecker
from tclint.violations import Violation
from tclint.comments import CommentVisitor
from tclint import utils

try:
    from tclint._version import __version__
except ModuleNotFoundError:
    __version__ = "(unknown version)"

# exit code flags
EXIT_OK = 0
EXIT_LINT_VIOLATIONS = 1
EXIT_SYNTAX_ERROR = 2
EXIT_INPUT_ERROR = 4


def resolve_sources(
    paths: List[pathlib.Path], exclude: List[pathlib.Path] = None
) -> List[pathlib.Path]:
    """Resolves paths passed via CLI to a list of filepaths to lint.

    `paths` is a list of paths that may be files or directories. Files are
    returned verbatim if they exist, and directories are recursively searched
    for files that have the extension .tcl, .xdc, or .sdc. Paths that match or
    are underneath a path provided in `exclude` are ignored.

    Raises FileNotFoundError if a supplied path does not exist.
    """
    # Extensions that may indicate tcl files
    # TODO: make configurable
    EXTENSIONS = [".tcl", ".xdc", ".sdc"]

    if exclude is None:
        exclude = []
    exclude = [path.resolve() for path in exclude]

    def is_excluded(path):
        resolved_path = path.resolve()
        for exclude_path in exclude:
            # if the current path is under an excluded path it should be ignored
            if utils.is_relative_to(resolved_path, exclude_path):
                return True

        return False

    sources = []

    for path in paths:
        if not path.exists():
            raise FileNotFoundError(f"path {path} does not exist")

        if is_excluded(path):
            continue

        if path.is_file():
            sources.append(path)
            continue

        assert path.is_dir()

        for dirpath, _, filenames in os.walk(path):
            for name in filenames:
                _, ext = os.path.splitext(name)
                if ext in EXTENSIONS:
                    child = pathlib.Path(dirpath) / name
                    if not is_excluded(child):
                        sources.append(child)

    return sources


def filter_violations(violations, config_ignore, inline_ignore, path):
    global_ignore = []
    path = path.resolve()
    for entry in config_ignore:
        if isinstance(entry, str):
            global_ignore.append(entry)
        else:
            ignore_path = entry["path"].resolve()
            if utils.is_relative_to(path, ignore_path):
                global_ignore.extend(entry["rules"])

    filtered_violations = []

    for violation in violations:
        if violation.id in global_ignore:
            continue
        line = violation.pos[0]
        if line in inline_ignore and violation.id in inline_ignore[line]:
            continue

        filtered_violations.append(violation)

    return filtered_violations


def lint(
    script: str, config: Config, path: pathlib.Path, debug=False
) -> List[Violation]:
    parser = Parser(debug=debug)

    violations = []
    tree = parser.parse(script)
    violations += parser.violations

    if debug:
        print(tree.pretty())

    for checker in (IndentLevelChecker(), SpacingChecker(), LineChecker()):
        violations += checker.check(script, tree, config)

    v = CommentVisitor()
    ignore_lines = v.run(tree, path)
    violations = filter_violations(violations, config.ignore, ignore_lines, path)

    return violations


def replace_with_warning_handler(e):
    # TODO: formal warning mechanism, include path
    print("Warning: non-unicode characters in file, replacing with �")
    return codecs.replace_errors(e)


codecs.register_error("replace_with_warning", replace_with_warning_handler)


def main():
    parser = argparse.ArgumentParser("tclint")
    parser.add_argument(
        "--version", action="version", version=f"%(prog)s {__version__}"
    )
    parser.add_argument("source", nargs="+", help="files to lint", type=pathlib.Path)
    parser.add_argument("--debug", action="store_true", help="display debug output")
    parser.add_argument(
        "-c", "--config", help="path to config file", type=pathlib.Path, default=None
    )
    args = parser.parse_args()

    try:
        config = get_config(args.config)
    except ConfigError as e:
        print(f"Invalid config file: {e}")
        return EXIT_INPUT_ERROR

    try:
        sources = resolve_sources(args.source, exclude=config.exclude)
    except FileNotFoundError as e:
        print(f"Invalid path provided: {e}")
        return EXIT_INPUT_ERROR

    retcode = EXIT_OK

    for path in sources:
        with open(path, "r", errors="replace_with_warning") as f:
            script = f.read()

        try:
            violations = lint(script, config, path, debug=args.debug)
        except TclSyntaxError as e:
            print(f"{path}: syntax error: {e}")
            retcode |= EXIT_SYNTAX_ERROR
            continue

        for violation in sorted(violations):
            print(f"{path}:{violation}")

        if len(violations) > 0:
            retcode |= EXIT_LINT_VIOLATIONS

    return retcode


if __name__ == "__main__":
    sys.exit(main())
